@@ -11,29 +11,19 @@
  */
 
 #include  "config.h"
+#include  "fetchmail.h"
+
 #include  <stdio.h>
 #include  <errno.h>
 #include  <string.h>
+#include  <strings.h>
 #include  <signal.h>
-#ifdef HAVE_MEMORY_H
-#include  <memory.h>
-#endif /* HAVE_MEMORY_H */
-#if defined(STDC_HEADERS)
 #include  <stdlib.h>
-#endif
-#if defined(HAVE_UNISTD_H)
 #include  <unistd.h>
-#endif
-#if defined(HAVE_STDARG_H)
 #include  <stdarg.h>
-#else
-#include  <varargs.h>
-#endif
 #include  <ctype.h>
 #include  <langinfo.h>
 
-#include  "fetchmail.h"
-
 /* for W* macros after pclose() */
 #define _USE_BSD
 #include <sys/types.h>
@@ -189,6 +179,16 @@ int smtp_setup(struct query *ctl)
 	    }
 	    else
 	    {
+		if (ctl->server.esmtp_name || ctl->server.esmtp_password)
+		{
+		    // we have an ESMTP user name or password, but
+		    // old SMTP servers that don't understand EHLO
+		    // cannot negotiate AUTH SASL mechanisms, so we need
+		    // to skip
+		    xfree(parsed_host);
+		    continue;
+		}
+
 		if ((ctl->smtp_socket = SockOpen(parsed_host,portnum,
 				ctl->server.plugout, &ai1)) == -1)
 		{
@@ -245,7 +245,8 @@ static void sanitize(char *s)
 char *rcpt_address(struct query *ctl, const char *id,
 			  int usesmtpname)
 {
-    static char addr[HOSTLEN+USERNAMELEN+1];
+    static char addr[HOSTLEN+USERNAMELEN+1000];
+
     if (strchr(id, '@'))
     {
 	snprintf(addr, sizeof (addr), "%s", id);
@@ -444,18 +445,6 @@ static int handle_smtp_report(struct query *ctl, struct msgblk *msg)
 
     responses[0] = xstrdup(smtp_response);
 
-#ifdef __UNUSED__
-    /*
-     * Don't do this!  It can really mess you up if, for example, you're
-     * reporting an error with a single RCPT TO address among several;
-     * RSET discards the message body and it doesn't get sent to the
-     * valid recipients.
-     */
-    smtp_rset(ctl);    /* stay on the safe side */
-    if (outlevel >= O_DEBUG)
-	report(stdout, GT_("Saved error is still %d\n"), smtperr);
-#endif /* __UNUSED */
-
     /*
      * Note: send_bouncemail message strings are not made subject
      * to gettext translation because (a) they're going to be 
@@ -532,12 +521,6 @@ static int handle_smtp_report(struct query *ctl, struct msgblk *msg)
 	 * (b) we wouldn't want spammers to get confirmation that
 	 * this address is live, anyway.
 	 */
-#ifdef __DONT_FEED_THE_SPAMMERS__
-	if (run.bouncemail)
-	    send_bouncemail(ctl, msg, XMIT_ACCEPT,
-			"Invalid address in MAIL FROM (SMTP error 553).\r\n", 
-			1, responses);
-#endif /* __DONT_FEED_THE_SPAMMERS__ */
 	free(responses[0]);
 	return(PS_REFUSED);
 
@@ -616,10 +599,6 @@ static int handle_smtp_report_without_bounce(struct query *ctl, struct msgblk *m
 	return(PS_REFUSED);
 
     case 553: /* invalid sending domain */
-#ifdef __DONT_FEED_THE_SPAMMERS__
-	if (run.bouncemail)
-	    return(PS_SUCCESS);
-#endif /* __DONT_FEED_THE_SPAMMERS__ */
 	return(PS_REFUSED);
 
     default:
@@ -844,7 +823,7 @@ static int open_smtp_sink(struct query *ctl, struct msgblk *msg,
     const char	*ap;
     struct	idlist *idp;
     char		options[MSGBUFSIZE]; 
-    char		addr[HOSTLEN+USERNAMELEN+1];
+    char		addr[HOSTLEN+USERNAMELEN+1000];
 #ifdef EXPLICIT_BOUNCE_ON_BAD_ADDRESS
     char		**from_responses;
 #endif /* EXPLICIT_BOUNCE_ON_BAD_ADDRESS */
@@ -1096,9 +1075,7 @@ static int open_mda_sink(struct query *ctl, struct msgblk *msg,
 	      int *good_addresses, int *bad_addresses)
 /* open a stream to a local MDA */
 {
-#ifdef HAVE_SETEUID
     uid_t orig_uid;
-#endif /* HAVE_SETEUID */
     struct	idlist *idp;
     int	length = 0, fromlen = 0, nameslen = 0;
     char	*names = NULL, *before, *after, *from = NULL;
@@ -1184,18 +1161,28 @@ static int open_mda_sink(struct query *ctl, struct msgblk *msg,
 	for (dp = after, sp = before; (*dp = *sp); dp++, sp++) {
 	    if (sp[0] != '%')	continue;
 
+	    if (sp > before && sp[-1] == '\'') {
+		report(stderr, GT_("MDA option contains single-quoted %%%c expansion.\n"), sp[1]);
+		report(stderr, GT_("Refusing to deliver. Check the manual and fix your mda option.\n"));
+		free(before);
+		free(after);
+		if (from) free(from);
+		if (names) free(names);
+		return PS_SYNTAX;
+	    }
+
 	    /* need to expand? BTW, no here overflow, because in
 	    ** the worst case (end of string) sp[1] == '\0' */
 	    if (sp[1] == 's' || sp[1] == 'T') {
 		*dp++ = '\'';
-		strcpy(dp, names);
+		if (names) strcpy(dp, names);
 		dp += nameslen;
 		*dp++ = '\'';
 		sp++;	/* position sp over [sT] */
 		dp--;	/* adjust dp */
 	    } else if (sp[1] == 'F') {
 		*dp++ = '\'';
-		strcpy(dp, from);
+		if (from) strcpy(dp, from);
 		dp += fromlen;
 		*dp++ = '\'';
 		sp++;	/* position sp over F */
@@ -1221,7 +1208,6 @@ static int open_mda_sink(struct query *ctl, struct msgblk *msg,
     if (outlevel >= O_DEBUG)
 	report(stdout, GT_("about to deliver with: %s\n"), before);
 
-#ifdef HAVE_SETEUID
     /*
      * Arrange to run with user's permissions if we're root.
      * This will initialize the ownership of any files the
@@ -1233,19 +1219,16 @@ static int open_mda_sink(struct query *ctl, struct msgblk *msg,
 	report(stderr, GT_("Cannot switch effective user id to %ld: %s\n"), (long)ctl->uid, strerror(errno));
 	return PS_IOERR;
     }
-#endif /* HAVE_SETEUID */
 
     sinkfp = popen(before, "w");
     free(before);
     before = NULL;
 
-#ifdef HAVE_SETEUID
     /* this will fail quietly if we didn't start as root */
     if (seteuid(orig_uid)) {
 	report(stderr, GT_("Cannot switch effective user id back to original %ld: %s\n"), (long)orig_uid, strerror(errno));
 	return PS_IOERR;
     }
-#endif /* HAVE_SETEUID */
 
     if (!sinkfp)
     {
@@ -1263,6 +1246,18 @@ static int open_mda_sink(struct query *ctl, struct msgblk *msg,
     return(PS_SUCCESS);
 }
 
+/** return a constant string that describes the sink type */
+const char* get_sink_type(const struct query *ctl)
+{
+    if (ctl->bsmtp) 	return GT_("BSMTP");
+    if (ctl->mda)	return GT_("MDA");
+    if (ctl->listener == LMTP_MODE)
+	    		return GT_("LMTP");
+    if (ctl->listener == SMTP_MODE)
+	    		return GT_("SMTP");
+    return GT_("sink"); // generic return value before smtp_setup
+}
+
 int open_sink(struct query *ctl, struct msgblk *msg,
 	      int *good_addresses, int *bad_addresses)
 /* set up sinkfp to be an input sink we can ship a message to */
@@ -1292,29 +1287,7 @@ int open_sink(struct query *ctl, struct msgblk *msg,
 	       ctl->smtphostmode,
 	       ctl->smtphost ? ctl->smtphost : "localhost");
 
-#ifndef FALLBACK_MDA
-	/* No fallback MDA declared.  Bail out. */
 	return(PS_SMTP);
-#else
-	/*
-	 * If user had things set up to forward offsite, no way
-	 * we want to deliver locally!
-	 */
-	if (ctl->smtphost && strcmp(ctl->smtphost, "localhost"))
-	    return(PS_SMTP);
-
-	/* 
-	 * User was delivering locally.  We have a fallback MDA.
-	 * Latch it in place, logging the error, and fall through.
-	 * Set stripcr as we would if MDA had been the initial transport
-	 */
-	ctl->mda = FALLBACK_MDA;
-	if (!ctl->forcecr)
-	    ctl->stripcr = TRUE;
-
-	report(stderr, GT_("can't raise the listener; falling back to %s"),
-			 FALLBACK_MDA);
-#endif
     }
 
     if (ctl->mda)		/* must deliver through an MDA */
@@ -1379,10 +1352,8 @@ int close_sink(struct query *ctl, struct msgblk *msg, flag forward)
 	/* close the delivery pipe, we'll reopen before next message */
 	if (sinkfp)
 	{
-	    if (ferror(sinkfp))
-		err = 1, e2 = errno;
-	    if ((fflush(sinkfp)))
-		err = 1, e2 = errno;
+	    if (ferror(sinkfp)) { err = 1; e2 = errno; }
+	    if (fflush(sinkfp)) { err = 1; e2 = errno; }
 
 	    errno = 0;
 	    rc = pclose(sinkfp);
@@ -1431,7 +1402,7 @@ int close_sink(struct query *ctl, struct msgblk *msg, flag forward)
 	    {
 		report(stderr, GT_("SMTP listener refused delivery\n"));
 	        smtp_rset(ctl);    /* stay on the safe side */
-		return(TRUE);
+		return !run.softbounce;
 	    }
 	}
 
@@ -1490,8 +1461,16 @@ int close_sink(struct query *ctl, struct msgblk *msg, flag forward)
 		    }
 		    if (smtp_err != SM_OK)
 		    {
-			responses[errors] = xstrdup(smtp_response);
-			errors++;
+			/*
+			 * amavis returns the SMTP code from the recieving
+			 * host after the DATA-DOT. So we need to compare the
+			 * response to the antispam option here instead.
+			 */
+			if (handle_smtp_report(ctl, msg) != PS_REFUSED) {
+			    /* Only count an error if the message was not refused */
+			    responses[errors] = xstrdup(smtp_response);
+			    errors++;
+			}
 		    }
 		}
 
@@ -1562,12 +1541,12 @@ int open_warning_by_mail(struct query *ctl)
     else				/* send to postmaster  */
 	status = open_sink(ctl, &reply, &good, &bad);
     if (status == 0) {
-	stuff_warning(NULL, ctl, "From: FETCHMAIL-DAEMON@%s",
+	stuff_warning(NULL, ctl, "From: ", "FETCHMAIL-DAEMON@%s",
 		ctl->smtpaddress ? ctl->smtpaddress : fetchmailhost);
-	stuff_warning(NULL, ctl, "Date: %s", rfc822timestamp());
-	stuff_warning(NULL, ctl, "MIME-Version: 1.0");
-	stuff_warning(NULL, ctl, "Content-Transfer-Encoding: 8bit");
-	stuff_warning(NULL, ctl, "Content-Type: text/plain; charset=\"%s\"", iana_charset);
+	stuff_warning(NULL, ctl, "Date: ", "%s", rfc822timestamp());
+	stuff_warning(NULL, ctl, "MIME-Version: ", "1.0");
+	stuff_warning(NULL, ctl, "Content-Transfer-Encoding: ", "8bit");
+	stuff_warning(NULL, ctl, "Content-Type: ", "text/plain; charset=\"%s\"", iana_charset);
     }
     return(status);
 }
@@ -1576,32 +1555,22 @@ int open_warning_by_mail(struct query *ctl)
 /* if rfc2047charset is non-NULL, encode the line (that is assumed to be
  * a header line) as per RFC-2047 using rfc2047charset as the character
  * set field */
-#if defined(HAVE_STDARG_H)
-void stuff_warning(const char *rfc2047charset, struct query *ctl, const char *fmt, ... )
-#else
-void stuff_warning(rfc2047charset, ctl, fmt, va_alist)
-const char *charset;
-struct query *ctl;
-const char *fmt;	/* printf-style format */
-va_dcl
-#endif
+void stuff_warning(const char *rfc2047charset, struct query *ctl, const char *pfx, const char *fmt, ... )
 {
     /* make huge -- i18n can bulk up error messages a lot */
-    char	buf[2*MSGBUFSIZE+4];
+    char	buf[3*MSGBUFSIZE+4];
     va_list ap;
 
+    snprintf(buf, sizeof(buf), "%s", pfx);
+
     /*
      * stuffline() requires its input to be writeable (for CR stripping),
      * so we needed to copy the message to a writeable buffer anyway in
      * case it was a string constant.  We make a virtue of that necessity
      * here by supporting stdargs/varargs.
      */
-#if defined(HAVE_STDARG_H)
     va_start(ap, fmt) ;
-#else
-    va_start(ap);
-#endif
-    vsnprintf(buf, sizeof(buf) - 2, fmt, ap);
+    vsnprintf(buf+strlen(buf), sizeof(buf)-strlen(buf)-2, fmt, ap);
     va_end(ap);
 
     snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), "\r\n");
@@ -1617,7 +1586,7 @@ va_dcl
 void close_warning_by_mail(struct query *ctl, struct msgblk *msg)
 /* sign and send mailed warnings */
 {
-    stuff_warning(NULL, ctl, GT_("-- \nThe Fetchmail Daemon"));
+    stuff_warning(NULL, ctl, "-- \n", GT_("The Fetchmail Daemon"));
     close_sink(ctl, msg, TRUE);
 }
 
